# flake8: noqa

import os
import webbrowser

import markdown2
from jinja2 import Environment, FileSystemLoader


def markdown_filter(value):
    return markdown2.markdown(value, extras=["fenced-code-blocks"])


def generate_html_report(models, output_path, missing_metadata: list[str]):
    template_path = os.path.join(os.path.dirname(__file__), "templates", "report_template.html")
    env = Environment(loader=FileSystemLoader(os.path.dirname(template_path)))
    env.filters["markdown"] = markdown_filter
    template = env.get_template(os.path.basename(template_path))

    rendered_report = template.render(models=models, missing_metadata=missing_metadata)
    with open(output_path, "w") as f:
        f.write(rendered_report)

    # Open the report in a new browser tab
    webbrowser.open(output_path)



# flake8: noqa

import glob
import html
import os
import re


def find_yaml_files(dbt_project_path: str):
    yaml_files = glob.glob(os.path.join(dbt_project_path, "**/*.yml"), recursive=True)
    yaml_files.extend(glob.glob(os.path.join(dbt_project_path, "**/*.yaml"), recursive=True))
    return yaml_files


def format_suggestion(suggestion: str):
    html_suggestion = ""

    escaped_suggestion = html.escape(suggestion)
    formatted_suggestion = escaped_suggestion.replace("\n", "<br>")
    html_suggestion = formatted_suggestion

    return html_suggestion



import argparse
import os

from dbt_ai.dbt import DbtModelProcessor
from dbt_ai.report import generate_html_report


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Generate improvement suggestions and check metadata coverage for dbt models"
    )
    parser.add_argument("-f", "--dbt-project-path", help="Path to the dbt project directory")
    parser.add_argument(
        "--create-models",
        help="Create dbt models using the provided prompt",
        default=None,
    )
    parser.add_argument(
        "-a",
        "--advanced-rec",
        action="store_true",
        help="Generate only advanced recommendations for dbt models",
    )
    parser.add_argument(
        "-d",
        "--database",
        help="Specify the type of database system the dbt project is built on",
        choices=["snowflake", "postgres", "redshift", "bigquery"],
        default="snowflake",
    )
    args = parser.parse_args()

    if not args.create_models:
        processor = DbtModelProcessor(args.dbt_project_path, args.database)

        models, missing_metadata = processor.process_dbt_models(advanced=args.advanced_rec)

        output_path = os.path.join(args.dbt_project_path, "dbt_model_suggestions.html")

        lineage_description, graph = processor.generate_lineage(models)
        # processor.plot_directed_graph(graph)

        print(f"Lineage description:\n {lineage_description}")

        generate_html_report(models, output_path, missing_metadata)
        advancedprint = "advanced " if args.advanced_rec else ""
        print(f"Generated {advancedprint}improvement suggestions report at: {output_path}")

        models_without_metadata = [model["model_name"] for model in models if not model["metadata_exists"]]

        if models_without_metadata:
            print("\nThe following models are missing metadata:")
            for model_name in models_without_metadata:
                print(f"  - {model_name}")
        else:
            print("\nAll models have associated metadata.")
    else:
        processor = DbtModelProcessor(args.dbt_project_path, args.database)
        prompt = args.create_models
        processor.create_dbt_models(prompt)


if __name__ == "__main__":
    main()



# flake8: noqa

import openai
import os
import requests


def generate_response(prompt) -> list:
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[
            {
                "role": "user",
                "content": """You are a helpful assistant that suggests only very basic improvements to dbt models based on the model content provided to you. Apply the rules outlined below. 
                I will provide further questions and the contents of the dbt model in a following message. Do not deviate from the rules listed below \
                Assume you are making suggestions to a very new data engineer who is new to dbt and maybe even SQL.
                Your suggestions should help ensure this new engineer is most effectively using dbt
                More rules: \
                - Do not provide suggestions regarding capturing metadata in a yml file, because this information \
                  is being provided as part of a separate check in this application \
                - Do NOT suggest writing comments in models \
                - Do NOT suggest using LIMIT if the model is already selecting a small number of records (e.g. under 1000) \
                - Do NOT suggest using JOIN to filter records if the model is already selecting a small number of records (e.g. under 1000) \
                - Do NOT suggest to consider adding a comment at the top of the model to explain the purpose of the query and any relevant context
                - Avoid providing too many conditional suggestions such as "If this table is big, then do this"
                - If you find or say that there are no  recommendations to provide, then do not proceed any further to provide anything else. Maybe add a compliment if it's nicely written!
                - Limit to 4 suggestions maximum \
                    \
            Formatting: 
            Suggestions for model `model_name`: \n\n
                - suggestion 1 \n
                - suggestion 2 \n
                - suggestion 3 \n
                """,
            },
            {"role": "user", "content": prompt},
        ],
        max_tokens=1024,
        n=1,
        stop=None,
        temperature=0.1,
    )
    return response.choices[0].message["content"].strip()


def generate_response_advanced(prompt) -> list:
    example_advanced_recommendations = """
        - Consider using a window function to calculate rolling averages or cumulative sums for a large dataset. This can improve query performance by reducing the need to perform multiple passes over the same data.
        - If a model contains a large number of columns, consider splitting it into multiple models to improve query performance and maintainability.
        - Consider using a common table expression (CTE) to break down a complex query into smaller, more manageable pieces. This can improve query readability and maintainability.
        - If a model contains a large amount of data, consider using a partitioning key to improve query performance. This can help distribute the data across multiple nodes and reduce the amount of data that needs to be scanned.
        - Consider using a temporary table to store intermediate results for a complex query. This can help simplify the query logic and improve query performance by reducing the need to repeat certain calculations.
        - If a model contains a large number of joins, consider using a star schema or a snowflake schema to simplify the data model and improve query performance.
    """
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[
            {
                "role": "user",
                "content": f"""You are a helpful assistant that suggests only very advanced improvements to dbt models based on the model content provided to you. Apply the rules outlined below. 
                I will provide further questions and the contents of the dbt model in a following message. Do not deviate from the rules listed below \
                Assume you are making suggestions to a highly skilled data engineer with strong knowledge of dbt and SQL.  \
                More rules: \
                - Avoid providing too many conditional suggestions such as "If this table is big, then do this"
                - If you find or say that there are no advanced recommendations to provide, then do not proceed any further to provide non-advanced recommendations. Maybe add a compliment if it's nicely written!
                - Limit to 4 suggestions maximum 
                - Here are a number of example recommendations that can be classified as advanced. Your recommendations should classify equally as advanced as these recommendations (but do not need to be the same):
                    {example_advanced_recommendations}
                    
            Formatting: 
            Suggestions for model `model_name`: \n\n
                - suggestion 1 \n
                - suggestion 2 \n
                - suggestion 3 \n
                """,
            },
            {"role": "user", "content": prompt},
        ],
        max_tokens=1024,
        n=1,
        stop=None,
        temperature=0,
    )
    return response.choices[0].message["content"].strip()


def generate_dalle_image(prompt: str, image_size: str = "1024x1024"):
    final_prompt = f"Draw a set of connected balls representing the nodes and edges of the following graph description: \
                    {prompt} \
                    "
    print(f"Generating AI image using DALL-E with the following prompt: {final_prompt}")
    response = openai.Image.create(
        prompt=prompt,
        n=1,
        size=image_size,
    )

    image_url = response.data[0].url.strip()
    image_binary = requests.get(image_url).content

    return image_binary


def generate_models(prompt: str, sources_yml: str) -> list[str]:
    # Combine prompt and sources.yml content
    prompt_with_sources = f"{prompt}\n\nSources YAML:\n\n{sources_yml}\n\n"
    output_dir = "models"
    # Generate response using OpenAI API
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that will write dbt models based on the provided prompt. The prompt includes useful information such as the contents of the sources.yml file. \
                If the logic needs to be split into multiple dbt models, please delimit the contents of each model with '===', which will be used to split the data to write into separate sql files later. \
                Do not put any explanations. Each model content should be divided with a === so that splitting by === would divide the models. The first line in the split model should have a 'model_name: modelname' with the actual model name. \
                The following lines after the model name should be the sql content with NO codeblock syntax. The last line of that model file should be the line prior to the next === \
                The user will likely provide enough information such as any join requirements, or aggregation requirements so use this information to correctly structure your dbt model queries.",
            },
            {"role": "user", "content": prompt_with_sources},
        ],
        max_tokens=400,
        n=1,
        stop=None,
        temperature=0,
    )

    # Extract the models from the response
    models = response.choices[0].message["content"].split("MODEL:")
    # print(models)
    return models



# flake8: noqa

import glob
import os
import re
from typing import Callable

import networkx as nx
import plotly.graph_objects as go
import yaml

from dbt_ai.ai import generate_dalle_image, generate_models, generate_response, generate_response_advanced
from dbt_ai.helper import find_yaml_files


class DbtModelProcessor:
    """Class containing functions to process and analyse a DBT project"""

    def __init__(self, dbt_project_path: str, database: str = "snowflake") -> None:
        self.dbt_project_path = dbt_project_path
        self.yaml_files = find_yaml_files(dbt_project_path)
        self.api_key_available = bool(os.getenv("OPENAI_API_KEY"))
        self.sources_yml_content = self.read_sources_yml(dbt_project_path)
        self.database = database
        if not self.api_key_available:
            print("Warning: OPENAI_API_KEY is not set. Suggestion features will be unavailable.")

    def read_sources_yml(self, dbt_project_path: str):
        sources_yml_path = os.path.join(dbt_project_path, "models", "sources.yml")
        try:
            with open(sources_yml_path, "r") as f:
                sources_yml_content = f.read()
        except FileNotFoundError:
            print("sources.yml not found. Proceeding with an empty sources file.")
            sources_yml_content = None
        return sources_yml_content

    def get_model_refs(self, model_file_path: str) -> list:
        with open(model_file_path, "r") as f:
            content = f.read()

        refs = re.findall(r"ref\(['\"]([\w\.]+)['\"]\)", content)

        return refs

    def suggest_dbt_model_improvements(self, file_path: str, model_name: str) -> list:
        with open(file_path, "r") as f:
            content = f.read()
        prompt = f"""Given the following dbt model {model_name}:\n\n{content}\n\nPlease provide suggestions on how to improve this model in terms of syntax, code structure and dbt best practices \
            such as using ref instead of hardcoding table names. The suggestion should be specific to db models written in the {self.database} database system: \
            """
        response = generate_response(prompt)
        return response

    def suggest_dbt_model_improvements_advanced(self, file_path: str, model_name: str) -> list:
        with open(file_path, "r") as f:
            content = f.read()
        prompt = f"""Given the following dbt model {model_name}:\n\n{content}\n\n \
            Please provide advanced suggestions on how to improve this model.
            The suggestion should be specific to dbt models written in the {self.database} database system 
            If there are no advanced recommendations to provide, then do not provide anything. If you are lacking context required to provide any advanced
            recommendations then don't provide anything. Example of an advanced recommendation include suggesting Snowflake partitioning keys when you see table names 
            being used that are very likely to be large tables e.g. invoice or journal line tables. Note that is just one example.
                """
        response = generate_response_advanced(prompt)
        return response

    def process_model(self, model_file: str, advanced: bool = False):
        model_name = os.path.basename(model_file).replace(".sql", "")

        has_metadata = self.model_has_metadata(model_name)
        if self.api_key_available:
            if advanced:
                raw_suggestion = self.suggest_dbt_model_improvements_advanced(model_file, model_name)
            else:
                raw_suggestion = self.suggest_dbt_model_improvements(model_file, model_name)
        else:
            raw_suggestion = ""

        refs = self.get_model_refs(model_file)

        return {
            "model_name": model_name,
            "metadata_exists": has_metadata,
            "suggestions": raw_suggestion,
            "refs": refs,
        }

    def process_dbt_models(self, advanced: bool = False):
        model_files = glob.glob(os.path.join(self.dbt_project_path, "models/**/*.sql"), recursive=True)
        models = [self.process_model(model_file, advanced) for model_file in model_files]
        missing_metadata = []

        # Check for models without metadata
        for model in models:
            if not model["metadata_exists"]:
                missing_metadata.append(model["model_name"])

        return models, missing_metadata

    def model_has_metadata(self, model_name: str) -> bool:
        for yaml_file in self.yaml_files:
            with open(yaml_file, "r") as f:
                try:
                    yaml_content = yaml.safe_load(f)

                    if yaml_content:
                        for item in yaml_content.get("models", []):
                            if isinstance(item, dict) and item.get("name") == model_name:
                                return True
                except yaml.YAMLError as e:
                    print(f"Error parsing YAML file {yaml_file}: {e}")

        return False

    def generate_lineage_graph(self, models):
        # Create a directed graph
        gph = nx.DiGraph()

        # Add nodes for each model
        for model in models:
            gph.add_node(model["model_name"], metadata_exists=model["metadata_exists"])

        # Add edges based on ref() calls
        for model in models:
            refs = model["refs"]
            for ref in refs:
                gph.add_edge(ref, model["model_name"])

        return gph

    def generate_lineage_description(self, gph: nx.DiGraph) -> str:
        nodes = list(nx.topological_sort(gph))

        description = ""  # "The following DBT models are used:\n\n"
        for node in nodes:
            parents = list(gph.predecessors(node))
            if parents:
                parent_names = ", ".join(parents)
                description += f"{node} depends on {parent_names}\n"
            else:
                description += f"{node} is a root node\n"

        return description

    def generate_lineage(self, dbt_models: list[dict]):
        gph = self.generate_lineage_graph(dbt_models)
        description = self.generate_lineage_description(gph)
        return description, gph

    def generate_image(self, description: str) -> None:
        image_binary = generate_dalle_image(description)
        image_path = f"{self.dbt_project_path}/lineage.png"
        print(f"Saving generated lineage image in {image_path}")
        # Write image to file
        with open(image_path, "wb") as f:
            f.write(image_binary)

    def plot_directed_graph(self, gph: nx.DiGraph):
        pos = nx.spring_layout(gph, seed=42)

        node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers+text",
            textposition="top center",
            hoverinfo="text",
            marker=dict(color="rgb(71, 122, 193)", size=10, line=dict(width=2, color="rgb(0, 0, 0)")),
            name="Nodes",
        )

        missing_metadata_color = "rgb(255, 0, 0)"

        edge_trace = go.Scatter(
            x=[],
            y=[],
            line=dict(width=2, color="#888"),
            hoverinfo="none",
            mode="lines",
            name="Edges",
        )

        for node in gph.nodes():
            x, y = pos[node]
            node_trace["x"] += tuple([x])
            node_trace["y"] += tuple([y])
            node_trace["text"] += tuple([node])

            # set marker color and/or text label based on whether the model has metadata or not
            if "metadata_exists" in gph.nodes[node] and not gph.nodes[node]["metadata_exists"]:
                node_trace["marker"]["color"] = missing_metadata_color
                node_trace["text"] += tuple(["MISSING METADATA"])
            else:
                node_trace["marker"]["color"] = "rgb(71, 122, 193)"

        for edge in gph.edges():
            x0, y0 = pos[edge[0]]
            x1, y1 = pos[edge[1]]
            edge_trace["x"] += tuple([x0, x1, None])
            edge_trace["y"] += tuple([y0, y1, None])

        fig = go.Figure(data=[edge_trace, node_trace])
        fig.update_layout(
            title="Directed Graph of DBT Models",
            title_x=0.5,
            title_font_size=24,
            showlegend=False,
            hovermode="closest",
            margin=dict(b=20, l=5, r=5, t=40),
            xaxis=dict(showgrid=False, zeroline=False, showticklabels=False, range=[-1.2, 1.2]),
            yaxis=dict(showgrid=False, zeroline=False, showticklabels=False, range=[-1.2, 1.2]),
        )

        fig.show()

    def create_dbt_models(self, prompt: str) -> None:
        print("Attempting to create dbt models based on prompt")
        sources_yml = self.sources_yml_content if self.sources_yml_content else ""
        response = generate_models(prompt, sources_yml)

        model_delimiter = "===\n\n"
        response_lines = response[0].split(model_delimiter)

        for _i, model_str in enumerate(response_lines):
            if not model_str.strip():
                continue

            model_lines = model_str.strip().split("\n")
            model_name = model_lines[0].split(":")[-1].strip()
            model_content = "\n".join(model_lines[1:])
            model_content = model_content.replace("===", "")

            model_path = os.path.join(self.dbt_project_path, "models", f"{model_name}.sql")
            with open(model_path, "w") as f:
                f.write(model_content.strip())
            print(f"Created model file: {model_path}")
